//
// MyBitmap.cpp
// ------------
// 
// DSgf : http://sourceforge.net/projects/dsgf
//

/* system includes */
#include <nds.h>
#include <stdlib.h>	// pour l'allocation mmoire malloc/free
#include <nds/arm9/console.h> // module console	pour iprintf (debug)
#include <stdio.h>						// fonctions standarts d'I/O (idem:debug/console)

/* my includes */
#include "MyTools.h"
#include "MyBitmap.h"

extern u16* pMAIN_BG;
extern u16* pMAIN_LARGE_BG;

extern CMyBitmap gobanNormal;
extern CMyBitmap gobanZoom;

// CMyBitmap

// Classe pour les bitmaps pr-dfinis (read_only=true) ou pour les drawables (crans virtuels et autres : read_only=false).

CMyBitmap::CMyBitmap(const u8* p_binary,u16 sx,u16 sy, bool transp,u8 coul_transp) // Read_only, zuste un pointeur sur le binaire
{
	p_data=(u8*)p_binary;
	size_x=sx;
	size_y=sy;
	transparence=transp;
	couleur_transp=coul_transp;
	read_only=true;	
}

CMyBitmap::CMyBitmap(u16 sx,u16 sy, bool transp,u8 coul_transp)	// Drawable, on alloue la mmoire.
{
	p_data = (u8*)malloc(sizeof(u8)*sx*sy);
	size_x=sx;
	size_y=sy;
	transparence=transp;
	read_only=false;
	couleur_transp=coul_transp;
}

void CMyBitmap::Fill(u8 color)
{
	int i;
	if (read_only)
	{
		erreurFatale("CMyBitmap:SetPixel : bitmap READONLY!");
	}
	else
	{
		for (i=0;i<=(size_x*size_y);i++)
			p_data[i]=color;
	}
}

void CMyBitmap::Fill(MyRect rect,u8 color)
{
	DrawRect(rect.x,rect.y,rect.width,rect.height,color);
}


void CMyBitmap::Copy2VRAM(u16* pdest)	// Copie 1 pour 1 sur un cran
{
	// Si on a pas de transparence et si on travaille sur des multiples de 2
	// alors on peut utiliser dmaCopy
	dmaCopy(p_data, pdest , size_x*size_y);	// formellement size_x*size_y, mais ca n'a de sens que pour un cran de 256x192 (sinon dcalage)
}

void CMyBitmap::Copy2VRAMscroll256x192(u16* pdest, u16 vertical_scroll)	// copie avec decalage vertical (ex.: bitmap(256,2*192) pour un scrolling vertical
{																														// pour l'instant que pour les crans de type 256*192!
	dmaCopy((p_data+256*vertical_scroll), pdest , 256*192);
}

void CMyBitmap::Copy2VRAMscroll256x126(u16* pdest, u16 vertical_scroll)	// version spcialise pour le scrolling de la zone de commentaire de l'cran de stats	
{																		
	dmaCopy((p_data+256*vertical_scroll), pdest , 256*126);	
}

void CMyBitmap::DrawRect(u16 x,u16 y,u16 dx, u16 dy, u8 palette_index)
{
	//iprintf("DrawRect\n");
	int i,j;
	for (i=0;i<dx;i++)
		for (j=0;j<dy;j++)
		{
			//iprintf("%d,%d;",i,j);
			SetPixel(x+i,y+j,palette_index);
		}
}
void CMyBitmap::DrawFrameRect(MyRect m_rect, u8 palette_index)
{
	DrawFrameRect(m_rect.x,m_rect.y,m_rect.width,m_rect.height,palette_index);
}

void CMyBitmap::DrawFrameRect(u16 x,u16 y,u16 dx,u16 dy, u8 palette_index)
{
	//iprintf("index=%d;",palette_index);
	int i,j;
	for (i=0;i<dx;i++)
	{
		SetPixel(x+i,y,palette_index);
		SetPixel(x+i,y+dy,palette_index);
	}

	for (j=0;j<dy;j++)
	{
		SetPixel(x,y+j,palette_index);
		SetPixel(x+dx,y+j,palette_index);
	}
}

void CMyBitmap::SetPixel(u16 x,u16 y,u8 color)
{
		p_data[x+y*size_x]=color;
}

u8 CMyBitmap::GetPixel(u16 x,u16 y)
{		
		u32 pos=x+y*size_x;
		u8 res=p_data[pos];
		return (res);
}
	
void CMyBitmap::Copy2Bitmap(u16 x0,u16 y0, u16 width, u16 height, CMyBitmap* pbmpdest, u16 x1, u16 y1)
{	
	if (pbmpdest->read_only)
	{
			iprintf("CMyBitmap::Copy2Bitmap : destination READ_ONLY!?!\n");
	}
	
	int x,y;	
		
	for (x=0;x<width;x++)
		for (y=0;y<height;y++)		
		{	pbmpdest->p_data[(x1+x)+(y+y1)*(pbmpdest->size_x)] = p_data[(x+x0)+(y+y0)*(size_x)]; }		
}

void CMyBitmap::Copy2Bitmap(CMyBitmap* pbmpdest, u16 x1, u16 y1)
{	
	if (pbmpdest->read_only)
	{
			iprintf("CMyBitmap::Copy2Bitmap : destination READ_ONLY!?!\n");
	}
	
	int x,y;	
		
	for (x=0;x<size_x;x++)
		for (y=0;y<size_y;y++)		
		{	pbmpdest->p_data[(x1+x)+(y+y1)*(pbmpdest->size_x)] = p_data[(x)+(y)*(size_x)]; }		
}

void CMyBitmap::BitBltOnBitmap(CMyBitmap* pbmpdest, u16 x1, u16 y1)
{
	if (pbmpdest->read_only)
	{
			iprintf("CMyBitmap::Copy2Bitmap : destination READ_ONLY!?!\n");
	}
	
	int x,y;	
	u8 value;
		
	for (x=0;x<size_x;x++)
		for (y=0;y<size_y;y++)		
		{	
			value=GetPixel(x,y);
			if (value!=0)
				pbmpdest->p_data[(x1+x)+(y+y1)*(pbmpdest->size_x)] = value;			
		}		
}

CMyBitmap::~CMyBitmap()
{
	if (!read_only)
	{
		free(p_data);
	}
}

void CMyBitmap::CopyToMAIN_BG(u16 xdest,u16 ydest)
{	
	
	int x,y;
	u8 value;
	
	for (x=0;x<(size_x);x++)
		for (y=0;y<(size_y);y++)
		{
			value=GetPixel(x,y);
			if (value!=0)
			{			
				SetPixelMAIN_BG((xdest+x),(ydest+y),value);			
			}
		}	
}

void CMyBitmap::RemoveFromMAIN_BG(u16 xdest,u16 ydest)
{
	int x,y;
	u8 value;
	
	for (x=0;x<(size_x);x++)
		for (y=0;y<(size_y);y++)
		{			
			if (GetPixel(x,y)!=0)
			{
				value=gobanNormal.GetPixel((xdest+x),(ydest+y));			
				SetPixelMAIN_BG((xdest+x),(ydest+y),value);			
			}
		}
}

void CMyBitmap::CopyToMAIN_LARGE_BG(u16 xdest,u16 ydest)
{	
	
	int x,y;
	u8 value;
	
	for (x=0;x<(size_x);x++)
		for (y=0;y<(size_y);y++)
		{
			value=GetPixel(x,y);
			if (value!=0)
			{			
				SetPixelMAIN_LARGE_BG((xdest+x),(ydest+y),value);			
			}
		}	
}

void CMyBitmap::RemoveFromMAIN_LARGE_BG(u16 xdest,u16 ydest)
{
	int x,y;
	u8 value;
	
	for (x=0;x<(size_x);x++)
		for (y=0;y<(size_y);y++)
		{			
			if (GetPixel(x,y)!=0)
			{				
				value=gobanZoom.GetPixel((xdest+x),(ydest+y));			
				SetPixelMAIN_LARGE_BG((xdest+x),(ydest+y),value);			
			}
		}
}

void CMyBitmap::SetPixelMAIN_BG(u16 x,u16 y,u8 value)
{
	// est-ce que x est pair ?
	// si x est pair, posu8=x+y*size_x est pair, car size_x est pair !
	
	u32 pos;
	
	u16 value16;
			
	//iprintf("pixel : x=%d y=%d; ",x,y);
	
	if ((x % 2)==0)
	{
		// x pair
		
		pos=(x/2+(y*(256/2)));
		
		//iprintf("pos=%d ",pos);
		
		value16=pMAIN_BG[pos];
		
		// on veut changer le pixel de droite
		
		value16 = value16 & 0xFF00;	// RAZ  droite
		value16 = value16 | value;		
		
		pMAIN_BG[pos]=value16;
		
	}
	else
	{
		// x impair
		x--; // plus maintenant !
		pos=(x/2+(y*(256/2)));
		
		value16=pMAIN_BG[pos];		

		// on veut changer le pixel de gauche
		
		value16 = value16 & 0x00FF;	// RAZ  gauche
		value16 = value16 | (value<<8);
		
		pMAIN_BG[pos]=value16;		
	}		
}

void CMyBitmap::SetPixelMAIN_LARGE_BG(u16 x,u16 y,u8 value)
{
	// est-ce que x est pair ?
	// si x est pair, posu8=x+y*size_x est pair, car size_x est pair !
	
	u32 pos;
	
	u16 value16;
		
	if ((x % 2)==0)
	{
		// x pair
		
		pos=(x/2+(y*(512/2)));
		
		value16=pMAIN_LARGE_BG[pos];
		
		// on veut changer le pixel de droite
		
		value16 = value16 & 0xFF00;	// RAZ  droite
		value16 = value16 | value;		
		
		pMAIN_LARGE_BG[pos]=value16;		
	}
	else
	{
		// x impair
		x--; // plus maintenant !
		pos=(x/2+(y*(512/2)));
		
		value16=pMAIN_LARGE_BG[pos];		

		// on veut changer le pixel de gauche
		
		value16 = value16 & 0x00FF;	// RAZ  gauche
		value16 = value16 | (value<<8);
		
		pMAIN_LARGE_BG[pos]=value16;	
	}		
}

void CMyBitmap::DrawHorizLine(u16 x,u16 y, u16 width, u8 color)
{
	u8 t;
	for (t=0;t<width;t++)
		SetPixel(t+x,y,color);
}
