//
// MySGF2.cpp
// ----------
// 
// Prise en charge du format SGF et des parties
//
// DSgf : http://sourceforge.net/projects/dsgf
//

//------------------------------------------------------------------------------

#include <nds.h>
#include <nds/arm9/console.h> // module console	pour iprintf (debug)
#include <stdio.h>			  // fonctions standarts d'I/O (idem:debug/console)

#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include <vector>
#include <stack>

#include <unistd.h>
#include <malloc.h>

#include "main.h"
#include "MyTools.h"
#include "MyGoban.h"

#include "MyBitmap.h"	// pour valuer  l'avance le nombre de pages de commentaires ncessaires
#include "MyFont.h"		// pour chaque commentaire de node
#include "MyLocalization.h"

#include "MySGF2.h"			// Structures et fonctions utilises pour grer l'arbre SGF

extern u8 __end__[];        // end of static code and data
extern u8 __eheap_end[];    // farthest point to which the heap will grow

u8* getHeapStart() {
   return __end__;
}
u8* getHeapEnd() {
   return (u8*)sbrk(0);
}
u8* getHeapLimit() {
   return __eheap_end;
}
int getMemUsed() {
   struct mallinfo mi = mallinfo();
   return mi.uordblks;
}
int getMemFree() {
   struct mallinfo mi = mallinfo();
   return mi.fordblks + (getHeapLimit() - getHeapEnd());
} 

//------------------------------------------------------------------------------

// Variables principales dfinies dans main.cpp

extern CMySGFTree	SGF_tree;	// l'arbre SGF

extern Goban_type	Goban;		// le goban

extern Coordonnees_type	Liste_capturees_g1[400];	// il peut y avoir jusqu' 4 groupes de capturs simultanment !
extern u16					Nb_capturees_g1;
extern bool				groupe1_vivant;

extern u16					Total_prisonniers_blancs;
extern u16					Total_prisonniers_noirs;

extern Pierre_type	Couleur_capturees;
extern u8 repartition_groupes [20][20];	// sert  dfinir les groupes de capture

extern u16					Nb_capturees_g2;
extern u16					Nb_capturees_g3;
extern u16					Nb_capturees_g4;

extern u16					Nombre_total_coups;

extern	CMyFont*	arial12;

extern CMyBitmap page_commentaire;

extern u16* pSUB_BG;

extern Localization	MyLocalization; // celle de main.cpp

//------------------------------------------------------------------------------

CMySGFTree::CMySGFTree()
{
	// constructeur
	allocated_node_number=0;
	allocated_goban_number=0;
	premiere_node=NULL;
}

CMySGFTree::~CMySGFTree()
{
	// destructeur	
	RAZ();
}

void CMySGFTree::ExtractHeaderAttributes(string entete)
{
	ExtractAttribute(entete,"AP",application);
	ExtractAttribute(entete,"CA",charset);
	ExtractAttribute(entete,"FF",fileformat);
	ExtractAttribute(entete,"GM",game);
	ExtractAttribute(entete,"ST",style);
	ExtractAttribute(entete,"SZ",size);
	ExtractAttribute(entete,"AN",annotation);
	ExtractAttribute(entete,"BR",black_rank);
	ExtractAttribute(entete,"BT",black_team);
	ExtractAttribute(entete,"CP",copyright);
	ExtractAttribute(entete,"DT",date);
	ExtractAttribute(entete,"EV",event);
	ExtractAttribute(entete,"GC",game_comment);
	ExtractAttribute(entete,"GN",game_name);
	ExtractAttribute(entete,"ON",opening);
	ExtractAttribute(entete,"PB",player_black);
	ExtractAttribute(entete,"PC",place);
	ExtractAttribute(entete,"PW",player_white);
	ExtractAttribute(entete,"RE",result);
	ExtractAttribute(entete,"RO",round);
	ExtractAttribute(entete,"RU",rules);
	ExtractAttribute(entete,"SO",source);
	ExtractAttribute(entete,"TM",time_limit);
	ExtractAttribute(entete,"US",user);
	ExtractAttribute(entete,"WR",white_rank);
	ExtractAttribute(entete,"WT",white_team);
	ExtractAttribute(entete,"HA",handicap);
	ExtractAttribute(entete,"KM",komi);
}

void CMySGFTree::ExtractAttribute(string source,string pattern,string &res)
{
	string::size_type p0,p1;
	p0=source.find(pattern+"[",0);
	
	if (p0==string::npos)
		res="";
	else
		{
			p0+=pattern.length()+1;	//p0 est sur le dbut de l'attribut
			p1=source.find("]",p0);
			while (source[p1-1]=='\\')	// juste pour viter les problmes avec des '\]' dans les attributs...
			{
				p1=source.find("]",p1+1);
			}
			res=source.substr(p0,p1-p0);
		}
}

//------------------------------------------------------------------------------

void CMySGFTree::remplis_arbre(string filename,CMyBitmap* p_ecran)	// cr la structure de l'arbre avec les commandes SGF
										// & analyse ces commandes (setup, move...)
{	

	string texte_ecran;

  u32 m_free,m_used;

	u16 nodes_lues=0;

	Nombre_total_coups=0;

	Node_type* node_courante=NULL;
	Node_type* nouvelle_node;

	u8 niveau=0; // niveau d'imbrication des variantes
	
	//Node_type* PileVariantes[16];	// 16 niveaux d'imbrication devraient suffire... (sauf pour un dictionnaire de joseki!)
	
	//Goban_type	PileGoban[16];		// idem pour sauvegarder l'tat du goban dans la pile	
	
	stack<Node_type*> PileVariantes;
	stack<Goban_type*> PileGoban;
	
	Goban_type*	pgoban;
	
	u16 move_number=0;	

	string buffer;
	u8 level;	// est utilis pour dtecter si le ';' est dans un attribut ou s'il est sparateur

#ifndef EMU_VERSION	
	int ch;
#else
	char ch;
#endif

	// intialisation de l'arbre
	
	RAZ();
	
	//TODO: faire un RAZ de l'arbre...
	// supprimer les nodes jusqu' la premire
		
	//premiere_node=NULL;
	
	// ouverture et lecture progressive du fichier

	#ifdef EMU_VERSION
	
				string* pheader;
				
				if (filename=="/SGF/Debug1.sgf")							
					pheader = new string((const char*)debug_sgf_raw);
				else if (filename=="/SGF/Debug2.sgf")							
					pheader = new string((const char*)debug_sgf2_raw);
				else if (filename=="/SGF/Debug3.sgf")							
					pheader = new string((const char*)debug_sgf3_raw);
				else if (filename=="/SGF/Debug4.sgf")							
					pheader = new string((const char*)debug_sgf4_raw);
				else if (filename=="/SGF/firzen.sgf")							
					pheader = new string((const char*)firzen_sgf_raw);
				else if (filename=="/SGF/jules.sgf")							
					pheader = new string((const char*)jules_sgf_raw);
				else
					pheader = new string((const char*)debug_sgf_raw);
				
				stringstream inputstream;
	
				inputstream.str(*pheader);
	
				delete pheader;
	#else  				
				//ifstream inputstream( filename.c_str(), ios_base::binary  );
				//filename="test.sgf";
				FILE* inputfile = fopen (filename.c_str(), "rb");
				//FILE* inputfile = fopen ("test.sgf", "rb");
							
	#endif
	
	u32 totalsize=0;		
	u32 currentsize=0;
	
	#ifdef EMU_VERSION
			if ( inputstream ) // ce test choue si le stream n'est pas ouvert
	#else  
			if ( inputfile!=NULL )
	#endif
			{		
				//iprintf("d3");
				//WaitForTouch();

	// Taille totale du fichier SGF ?
	#ifdef EMU_VERSION
		totalsize=GetFileSize(inputstream);
	#else
		totalsize=GetFileSize(inputfile);
	#endif
	  	
		#ifdef EMU_VERSION
			//getline(inputstream, buffer,';');					  	
			buffer="";
			level=0;
			inputstream.read(&ch,1);
			currentsize++;
			//iprintf("%c",ch);
			while( !(inputstream.eof()) )
			{	
				if ((ch=='[')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un [ d'attribut
					level++;
				if ((ch==']')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un ] d'attribut
					level--;
				if ((ch==';')&&(level==0))
					break;
				buffer+=ch;
				inputstream.read(&ch,1);
				//iprintf("%c",ch);
				currentsize++;
			}			
		#else	
			buffer="";
			level=0;
			ch = getc( inputfile );	
			currentsize++;	
			while( ch != EOF )
			{	
				if ((ch=='[')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un [ d'attribut
					level++;
				if ((ch==']')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un ] d'attribut
					level--;
				if ((ch==';')&&(level==0))
					break;
				buffer+=ch;
				ch = getc( inputfile );
				currentsize++;
			}
		#endif
		
			// entete ne contient ici que la parenthse d'ouverture du fichier SGF		
			
		#ifdef EMU_VERSION
			//getline(inputstream, buffer,';');					  	
			buffer="";
			level=0;
			inputstream.read(&ch,1);
			currentsize++;
			//iprintf("%c",ch);
			while( !(inputstream.eof()) )
			{	
				if ((ch=='[')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un [ d'attribut
					level++;
				if ((ch==']')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un ] d'attribut
					level--;
				if ((ch==';')&&(level==0))
					break;
				buffer+=ch;
				inputstream.read(&ch,1);
				currentsize++;
				//iprintf("%c",ch);
			}			
		#else	
			buffer="";
			level=0;
			ch = getc( inputfile );
			currentsize++;		
			while( ch != EOF )
			{	
				if ((ch=='[')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un [ d'attribut
					level++;
				if ((ch==']')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un ] d'attribut
					level--;
				if ((ch==';')&&(level==0))
					break;
				buffer+=ch;
				ch = getc( inputfile );
				currentsize++;
			}
		#endif
		
//				iprintf("buffer=%s\n",buffer.c_str());
//				WaitForTouch();
				
		// entete contient maintenant l'entte du fichier SGF...				  	
		// on va utiliser ExtractHeaderAttributes pour obtenir les attributs...	
				
	//			iprintf("extract on\n");
	//			WaitForTouch();
				ExtractHeaderAttributes(buffer);
	//			iprintf("extract off\n");
	//			WaitForTouch();
			}	
			else
			{
					// Failure
					iprintf("Erreur fatale a l'ouverture de %s!\n",filename.c_str());	
					WaitForTouch();  
			}

	p_ecran->DrawFrameRect(32,128,255-64,6,255);	// cadre noir pour la barre d'avancement

	//string sizeinfo=to_string(totalsize)+"\n";
	
	//iprintf(sizeinfo.c_str());

	initialise_goban(&Goban,size);

	// cration de la premire node...
	//nouvelle_node=(Node_type*)malloc(sizeof(Node_type));
	nouvelle_node=new (Node_type);
	allocated_node_number++;
	nouvelle_node->setup_goban=NULL;

	// initialisation des attributs de la node "root"
	// entete contient tout ce qu'il faut...
	
//iprintf("extract comment on\n");
//WaitForTouch();

//iprintf("buffer=%s",buffer.c_str());
//WaitForTouch();

	nouvelle_node->nb_volets_commentaires=ExtractComment(buffer,nouvelle_node->commentaire);

//iprintf("extract comment off\n");
//WaitForTouch();
	
	//extrait_commande_SGF_typechar(tmp_str,"C[",&(nouvelle_node->commentaire));

	nouvelle_node->move_number=(move_number++);
	
	// est-ce qu'il y a un setup ?

	//p_str1=strstr(tmp_str,"AW[");
	//p_str2=strstr(tmp_str,"AB[");
	//p_str3=strstr(tmp_str,"AE[");	

//iprintf("d0");
//WaitForTouch();

	bool root_node=true;

	if ((buffer.find("AW[",0)!=string::npos)||(buffer.find("AB[",0)!=string::npos)||(buffer.find("AE[",0)!=string::npos))
	{

		nouvelle_node->type_node=node_setup;

		// analyser les AW,AB,AE et remplir le goban...
	
		// pour l'instant, on prend uniquement en compte le cas du handicap
		// => AB[] (i.e. AddBlack)
		
		string::size_type p0;
		string tmp;
		Coordonnees_type co;
		
		p0=buffer.find("AB[",0);
		if (p0==string::npos)
		{
			iprintf("ERREUR: SETUP dans node_root autre que AB-handicap!");
			while(1);
		}
		
		p0+=2;
		
		//on prend les coordonnes suivantes...
		while (buffer[p0]=='[')
		{
			p0++;
			tmp=buffer.substr(p0,2);
			
			//iprintf("setup1 : .%s.",tmp.c_str());
			//WaitForTouch();
			
			p0+=3;
			convertit_coord(tmp,co,Goban.size);
			Goban.intersection[co.x][co.y]=PierreNoire;
		}		
		
		// on sauvegarde le goban dans goban_setup...
		
		nouvelle_node->setup_goban=new Goban_type;
		allocated_goban_number++;
		copie_goban(&Goban,nouvelle_node->setup_goban);
		
		// et ensuite charger le buffer suivant...
		
		#ifdef EMU_VERSION
			//getline(inputstream, buffer,';');					  	
			buffer="";
			level=0;
			inputstream.read(&ch,1);
			currentsize++;
			//iprintf("%c",ch);
			while( !(inputstream.eof()) )
			{	
				if ((ch=='[')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un [ d'attribut
					level++;
				if ((ch==']')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un ] d'attribut
					level--;
				if ((ch==';')&&(level==0))
					break;
				buffer+=ch;
				inputstream.read(&ch,1);
				//iprintf("%c",ch);
				currentsize++;
			}			
		#else	
			buffer="";
			level=0;
			ch = getc( inputfile );
			currentsize++;		
			while( ch != EOF )
			{	
				if ((ch=='[')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un [ d'attribut
					level++;
				if ((ch==']')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un ] d'attribut
					level--;
				if ((ch==';')&&(level==0))
					break;
				buffer+=ch;
				ch = getc( inputfile );
				currentsize++;
			}
		#endif
		
//		iprintf("buffer=%s",buffer.c_str());
//		WaitForTouch();
		root_node=false;

	}
	else
	{
		nouvelle_node->type_node=node_move;
		// en fait il n'y a pas de coup permettant d'arriver l, car le goban est vide !
	}

	premiere_node=nouvelle_node;
	node_courante=nouvelle_node;
	node_courante->node_name=MyLocalization.first_node_name;
	
// MAJ de l'avancement
u8 avancement=div32(189*currentsize, totalsize);

//sizeinfo=to_string(avancement)+":"+to_string(currentsize)+"/"+to_string(totalsize)+"\n";
//iprintf(sizeinfo.c_str());
					
p_ecran->DrawRect(33,129,avancement,5,253);	// orange
if (avancement<189)
	p_ecran->DrawRect(33+avancement,129,253-avancement-64,5,254); // blanc

//texte_ecran=node_courante->node_name+"\n"+to_string(++nodes_lues)+" noeuds lus.";	
//p_ecran->DrawRect(8,40,240,24,254);
//arial12->WriteParagraph(texte_ecran,p_ecran,8,40,248,64);
swiWaitForVBlank();	
p_ecran->Copy2VRAM(pSUB_BG);

//iprintf("debut du parcours...\n\n\n");

/*m_free=getMemFree()/1024;
m_used=getMemUsed()/1024;
iprintf("%d Ko used / %d Ko	free\n",m_used,m_free);
WaitForTouch();			
*/
++nodes_lues;
	
	//////////////////////////////////////////////////////////////////////
	// on commence le parcours du fichier SGF et la cration de l'arbre...

	// ide retenue : on va lire le fichier par segment, en prenant ';' comme dlimiteur...
	
	// chaque segment va tre stock dans buffer
	
	// la premire partie contient les commandes SGF
	// la deuxime les ventuelles parenthses
		
	bool c_le_dernier_tour=false;
	
	string commandes;
	string parentheses;
	
	string::size_type p,p0,p1;	

	while (1)
	{
		// on dlimite la partie commandes de la partie parenthses...
		// les parenthses sont  la fin...
		
		p=buffer.size()-1;
		
		//il faut maintenant reculer tant que l'on a des parenthses...
		
		while ((buffer[p]==')')||(buffer[p]=='(')||(buffer[p]=='\n')||(buffer[p]=='\r'))
		{
			p--;
		}
		p++;
		commandes=buffer.substr(0,p);
		parentheses=buffer.substr(p);

//iprintf("comm=.%s.\npar=.%s.\n\n",commandes.c_str(),parentheses.c_str());
//WaitForTouch();

		if ((commandes!="")&&(!root_node)) // pas d'analyse si root_node, c'est l'entte!
		{
			// cration d'une nouvelle node...
			//nouvelle_node=(Node_type*)malloc(sizeof(Node_type));
			nouvelle_node=new (Node_type);
			allocated_node_number++;
			// initialisation des attributs			
			nouvelle_node->nb_volets_commentaires=0;
			nouvelle_node->commande_SGF="";
			nouvelle_node->setup_goban=NULL;
			nouvelle_node->move_number=(move_number++);			
			//nouvelle_node->setup_goban=NULL;
				
			// on ajoute nouvelle_node aux variantes de node_courante;
			
			node_courante->variantes.push_back(nouvelle_node);

			nouvelle_node->node_amont=node_courante;
			node_courante=nouvelle_node;

			// On lit la commande SGF...

//			node_courante->commande_SGF=commandes;	

//			iprintf("%d comm=%s\n",node_courante->move_number,commandes.c_str());
//			WaitForTouch();
		
			// la recherche de setup est vire car pas de setup hors commande niveau root! (simplification DSgf)

/*
			// Recherche d'ventuelles commandes de type setup (AW,AE,AB)

			// est-ce qu'il y a un setup ?

			p_str1=strstr(node_courante->commande_SGF,"AW[");
			p_str2=strstr(node_courante->commande_SGF,"AB[");
			p_str3=strstr(node_courante->commande_SGF,"AE[");

			if ((p_str1!=NULL)||(p_str1!=NULL)||(p_str1!=NULL))
			{

				node_courante->type_node=node_setup;

				// analyser les AW,AB,AE et remplir le goban...
				
				// rien pour l'instant...

				PA_OutputText(1, 0, 1, "remplis_arbre:ERROR_2!");
				plante();

			}
			else
			{
			*/
				// move
				// on joue un coup "normal"...
				
				node_courante->type_node=node_move;
				
				p0=commandes.find("W[");
				p1=commandes.find("B[");				
				
				if ((p0!=string::npos) && (p0==0) && (commandes[p0+2]==']'))	// W[] => W passes
				{					
					string tmp;
					ExtractAttribute(commandes,"W",tmp);
					
					
					node_courante->node_name="White : Pass";
					
					node_courante->type_node=node_pass;
					node_courante->move_couleur=PierreBlanche;
					Couleur_capturees=PierreNoire;	
					nouvelle_node->move_number--;
					move_number--;
				}
				else
				if ((p1!=string::npos) && (p1==0) && (commandes[p1+2]==']'))	// B[] => B passes
				{					
					string tmp;
					ExtractAttribute(commandes,"B",tmp);
						
					node_courante->node_name="Black : Pass";
					
					node_courante->type_node=node_pass;
					node_courante->move_couleur=PierreNoire;
					Couleur_capturees=PierreBlanche;					
					nouvelle_node->move_number--;
					move_number--;
				}
				else
				if (p0!=string::npos && ((p0==0)||(commandes[p0-1]==']')||(commandes[p0-1]=='\n')))	// white move
				{
					string tmp;
					ExtractAttribute(commandes,"W",tmp);
					
					convertit_coord(tmp,node_courante->move_coord,Goban.size);
					
					if (node_courante->move_coord.x>19 || node_courante->move_coord.y>19)	// it's a pass move
					{
						node_courante->type_node=node_pass;
						node_courante->node_name="White : Pass";
						nouvelle_node->move_number--;
						move_number--;
					}
					else
						Goban.intersection[node_courante->move_coord.x][node_courante->move_coord.y]=PierreBlanche;

					node_courante->move_couleur=PierreBlanche;
					Couleur_capturees=PierreNoire;	
				}
				else				
				if (p1!=string::npos && ((p1==0)||(commandes[p1-1]==']')||(commandes[p0-1]=='\n')))	// black move
				{
					string tmp;

					ExtractAttribute(commandes,"B",tmp);

					convertit_coord(tmp,node_courante->move_coord,Goban.size);
					
					if (node_courante->move_coord.x>19 || node_courante->move_coord.y>19)	// it's a pass move
					{
						node_courante->type_node=node_pass;
						node_courante->node_name="Black : Pass";
						nouvelle_node->move_number--;
						move_number--;
					}
					else
						Goban.intersection[node_courante->move_coord.x][node_courante->move_coord.y]=PierreNoire;
					
					node_courante->move_couleur=PierreNoire;
					Couleur_capturees=PierreBlanche;					
				}
				else
				{
					// pas de dplacement...
					
					nouvelle_node->move_number--;
					
					//iprintf("comm=.%s.\npar=.%s.\n\n",commandes.c_str(),parentheses.c_str());
					//WaitForTouch();
					
					// C'est soit une node de setup (pas support pour l'instant)
					
					// soit une node "neutre" sans setup ni move (avec marqueurs & commentaires, par ex.).
					
					p0=commandes.find("N[");	// node_name
					
					if (p0!=string::npos && ((p0==0)||(commandes[p0-1]==']')||(commandes[p0-1]=='\n')))
					{
						//iprintf("Nodename!\n");
						//WaitForTouch();
						node_courante->type_node=node_neutral;
						string tmp;
						ExtractAttribute(commandes,"N",tmp);
						node_courante->node_name=tmp;

						//p_ecran->DrawRect(8,40,240,12,254);
						
						// MAJ de l'avancement
						u8 avancement=div32(189*currentsize, totalsize);
						
/*						string sizeinfo=to_string((u32)avancement)+":"+to_string(currentsize)+"/"+to_string(totalsize)+"\n";
						iprintf(sizeinfo.c_str());
*/
					//++nodes_lues;	
					//iprintf("%d nodes lues.\n",nodes_lues);

											
						p_ecran->DrawRect(33,129,avancement,5,253);	// orange
						if (avancement<189)
							p_ecran->DrawRect(33+avancement,129,253-avancement-64,5,254); // blanc
						
						//texte_ecran=node_courante->node_name+"\n"+to_string(++nodes_lues)+" noeuds lus.";	
						//p_ecran->DrawRect(8,40,240,24,254);
						//arial12->WriteParagraph(texte_ecran,p_ecran,8,40,248,64);
						swiWaitForVBlank();	
						p_ecran->Copy2VRAM(pSUB_BG);
					}
					else
					{
						//iprintf("buffer=.%s.\n",buffer.c_str());
						//iprintf("comm=.%s.\npar=.%s.\n\n",commandes.c_str(),parentheses.c_str());
						//iprintf("node de setup en dehors du root!\n");
						//WaitForTouch();
						node_courante->type_node=node_setup;
						node_courante->setup_goban=NULL;
						node_courante->move_coord.x=0;
						node_courante->move_coord.x=0;	
						node_courante->node_name="setup (unsupported here!)";

						// MAJ de l'avancement
						u8 avancement=div32(189*currentsize, totalsize);						
						
            string sizeinfo=to_string((u32)avancement)+":"+to_string(currentsize)+"/"+to_string(totalsize)+"\n";
						//iprintf(sizeinfo.c_str());
            					
						p_ecran->DrawRect(33,129,avancement,5,253);	// orange
						if (avancement<189)
							p_ecran->DrawRect(33+avancement,129,253-avancement-64,5,254); // blanc

				++nodes_lues;	
				//iprintf("%d nodes lues.\n",nodes_lues);
						
						//texte_ecran=node_courante->node_name+"\n"+to_string(++nodes_lues)+" noeuds lus.";	
						//p_ecran->DrawRect(8,40,240,24,254);
						//arial12->WriteParagraph(texte_ecran,p_ecran,8,40,248,64);
						swiWaitForVBlank();	
						p_ecran->Copy2VRAM(pSUB_BG);
						
					}
					node_courante->nb_volets_commentaires=ExtractComment(buffer,node_courante->commentaire);
				}

				if (node_courante->type_node==node_move)
				{
				
					Nombre_total_coups++;
				
					// nommage de la node...
					string texte;
					
					if (node_courante->move_couleur==PierreBlanche)
						texte="W ";
					else
						texte="B ";
						
					texte+=gobanx2letter(node_courante->move_coord.x)+to_string((u16)node_courante->move_coord.y);
				
					if (node_courante->node_name=="")
						node_courante->node_name=texte;

					// MAJ de l'avancement
					u8 avancement=div32(189*currentsize, totalsize);
/*					
					string sizeinfo=to_string((u32)avancement)+":"+to_string(currentsize)+"/"+to_string(totalsize)+"\n";
					iprintf(sizeinfo.c_str());
*/					
//++nodes_lues;	
//iprintf("%d nodes lues.\n",nodes_lues);
										
					p_ecran->DrawRect(33,129,avancement,5,253);	// orange
					if (avancement<189)
						p_ecran->DrawRect(33+avancement,129,253-avancement-64,5,254); // blanc
					
					//texte_ecran=node_courante->node_name+"\n"+to_string(++nodes_lues)+" noeuds lus.";	
					//p_ecran->DrawRect(8,40,240,24,254);
					//arial12->WriteParagraph(texte_ecran,p_ecran,8,40,248,64);
					swiWaitForVBlank();	
					p_ecran->Copy2VRAM(pSUB_BG);
				
					// Extraction du commentaire potentiel...
					
					node_courante->nb_volets_commentaires=ExtractComment(buffer,node_courante->commentaire);
					
					//iprintf("buff=%s.\n",buffer.c_str());
					
//					iprintf("size_node=%d\n",sizeof(*node_courante)+sizeof(node_courante->commentaire));
					
					
					
          m_free=getMemFree()/1024;
          m_used=getMemUsed()/1024;
          //iprintf("%d Ko used / %d Ko	free\n",m_used,m_free);          				
					
					//arial12->WriteParagraph((char*)MyLocalization.loading_text.c_str(),pHaut,8,60,248,84);					
					
					if (m_free<=1)
					{
					 iprintf("SORRY! memory full: %d Ko used / %d Ko	free\nsystem halted.",m_used,m_free);
					 while(1)
					 {            
           }
          }
					
					//extrait_commande_SGF_typechar(node_courante->commande_SGF,"C[",&(node_courante->commentaire));
	
					// Analyse du coup pour dterminer s'il y a capture...
	
					// On travaille sur Goban (var. globale) donc no problem...
					// comme si on venait de jouer le coup en mode enregistrement...				
	
					verifie_capture(node_courante->move_coord.x,node_courante->move_coord.y);
	
					// le rsultat est dans les variables globales correspondantes
					
					if ((Nb_capturees_g1+Nb_capturees_g2+Nb_capturees_g3+Nb_capturees_g4)>0)	// capture(s)!
					{
						//iprintf("Capture!\n");
						//WaitForTouch();
						// on prpare le tableau qui va bien...
						//node_courante->move_captures=(Coordonnees_type*)malloc(node_courante->nb_captures*sizeof(Coordonnees_type));
	
						// on remplit  la main groupe par groupe !
						u16 i,j;
						Coordonnees_type p;
						//p=node_courante->move_captures;
	
						for (i=1;i<=Goban.size;i++)
						{
							for (j=1;j<=Goban.size;j++)
							{
								if (repartition_groupes[i][j]==1 && Nb_capturees_g1>0)
								{
									p.x=i;
									p.y=j;
									node_courante->move_captures.push_back(p);
									// elle est capture, on l'enlve...
									Goban.intersection[i][j]=AucunePierre;
								}
								else
								if (repartition_groupes[i][j]==2 && Nb_capturees_g2>0)
								{
									p.x=i;
									p.y=j;
									node_courante->move_captures.push_back(p);
									// elle est capture, on l'enlve...
									Goban.intersection[i][j]=AucunePierre;
								}
								else
								if (repartition_groupes[i][j]==3 && Nb_capturees_g3>0)
								{
									p.x=i;
									p.y=j;
									node_courante->move_captures.push_back(p);
									// elle est capture, on l'enlve...
									Goban.intersection[i][j]=AucunePierre;
								}
								else
								if (repartition_groupes[i][j]==4 && Nb_capturees_g4>0)
								{
									p.x=i;
									p.y=j;
									node_courante->move_captures.push_back(p);
									// elle est capture, on l'enlve...
									Goban.intersection[i][j]=AucunePierre;
								}
							}					
						}
					}
				}
			//}	
						
		}
		

		
		while (parentheses.length()>0)
		{	
		
			//iprintf("par=.%s.\n",parentheses.c_str());
			//WaitForTouch();
			
			string p=parentheses.substr(0,1);
			parentheses=parentheses.substr(1);
			if (p=="(")
			{				
				PileVariantes.push(node_courante);
				pgoban=new Goban_type;
				allocated_goban_number++;
				copie_goban(&Goban,pgoban);
				PileGoban.push(pgoban);
				niveau++;
				//copie_goban(&Goban,&(PileGoban[niveau++]));
				//iprintf("Empile : niveau=%d\n",niveau);
				//WaitForTouch();
			}
			else
			if (p==")")
			{
				if (niveau==0)
				{
					// on peut pas dpiler, banane ! Il n'y a rien dans la pile !
					// c'est la dernire parenthse du fichier, a...
				}
				else
				{
					niveau--;
					//iprintf("depile : niveau=%d\n",niveau);
					//WaitForTouch();
					//node_courante=PileVariantes[niveau];
					//copie_goban(&(PileGoban[niveau]),&Goban);
					
					node_courante=PileVariantes.top();
					PileVariantes.pop();
					pgoban=PileGoban.top();
					PileGoban.pop();
					copie_goban(pgoban,&Goban);
					delete pgoban;
					allocated_goban_number--;
					move_number=node_courante->move_number+1;	
				}
			}
		}	
		if (root_node)
			root_node=false;
		
		// on charge le segment suivant dans buffer...

		if (c_le_dernier_tour)
			break;

		//iprintf("avant getline\n");
		//WaitForTouch();
		
		#ifdef EMU_VERSION
			//getline(inputstream, buffer,';');					  	
			buffer="";
			level=0;
			inputstream.read(&ch,1);
			currentsize++;
			//iprintf("%c",ch);
			while( !(inputstream.eof()) )
			{	
				if ((ch=='[')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un [ d'attribut
					level++;
				if ((ch==']')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un ] d'attribut
					level--;
				if ((ch==';')&&(level==0))
					break;
				buffer+=ch;
				inputstream.read(&ch,1);
				//iprintf("%c",ch);
				currentsize++;
			}			
		#else	
			buffer="";
			level=0;
			ch = getc( inputfile );
			currentsize++;		
			while( ch != EOF )
			{	
				if ((ch=='[')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un [ d'attribut
					level++;
				if ((ch==']')&&(buffer!="")&&(buffer[buffer.size()-1]!='\\'))	// C'est bien un ] d'attribut
					level--;
				if ((ch==';')&&(level==0))
					break;
				buffer+=ch;
				ch = getc( inputfile );
				currentsize++;
			}
		#endif

//if (buffer.length()!=5)
//{
//	ClearConsole();
//	iprintf("buffer=.%s.\n",buffer.c_str());
//	WaitForTouch();
//}		
		//iprintf("buff=.%s.\n",buffer.c_str());
		//WaitForTouch();
		#ifdef EMU_VERSION
		if (inputstream.eof())
		#else 
		if (ch == EOF)
		#endif
		{				
			c_le_dernier_tour=true;
		}	
		//iprintf("nouveau buffer:%s\n",buffer.c_str());
		//WaitForTouch();
	}
	
	// fin de l'analyse...

	#ifndef EMU_VERSION
				
			//inputstream.close();
			fclose(inputfile);
					    
	#endif

	//iprintf("lecture terminee\n",buffer.c_str());
	//WaitForTouch();

	// vrifications...
	
	if (PileVariantes.size()!=0)
	{
		iprintf("Erreur! il reste %d elements dans PileVariantes!!!\n",PileVariantes.size());
		while(1);
	}

	if (PileGoban.size()!=0)
	{
		iprintf("Erreur! il reste %d elements dans PileGoban!!!\n",PileGoban.size());
		while(1);
	}

	initialise_goban(&Goban,size);	// RAZ du goban
}

s8 CMySGFTree::Move_forward(Node_type* node)	// return true si capture
{
	if (node->type_node==node_setup||node->type_node==node_neutral||node->type_node==node_pass)
	{
		//iprintf("Move_forward:setup ou neutral!\n");		
		//WaitForTouch();
		return -1;
	}	
	// on joue la pierre "qui va bien" et on capture la "capture_list" !
	Pierre_type pierre;
	if (node->move_couleur==PierreBlanche)
	{	pierre=PierreBlanche;
	}
	else
	{	pierre=PierreNoire;
	}

	//iprintf("on pose une pierre %i a %i,%i\n",pierre,node->move_coord.x,node->move_coord.y);
	pose_pierre(node->move_coord.x,node->move_coord.y,pierre);
/*
	// on pose le triangle
	pose_triangle_rouge(node->move_coord.x,node->move_coord.y);

	// on enlve l'ancien triangle
	Node_type* previous_node=node->node_amont;

	if (previous_node!=NULL)
	{
		pose_pierre(previous_node->move_coord.x,previous_node->move_coord.y,previous_node->move_couleur);
	}
*/
	if (node->move_captures.size()>0)
	{
		SGF_tree.Capture_pierres(node->move_captures);
		if (node->move_couleur==PierreBlanche)
		{
			Total_prisonniers_noirs+=node->move_captures.size();
		}
		else
		{
			Total_prisonniers_blancs+=node->move_captures.size();
		}
		return 1;
	} 
	else	// s'il y a capture, il ne peut y avoir suicide de la pierre !
	{
		// Maintenant on teste s'il y a suicide de la pierre pose => coup interdit!

		u8 a,b;

		Couleur_capturees=node->move_couleur;

		Nb_capturees_g1=0;

		groupe1_vivant=false;	// jusqu' preuve du contraire !

		for (a=1;a<=19;a++)
			for (b=1;b<=19;b++)
			{
				repartition_groupes[a][b]=0;
			}
			
		// Affectation du groupe1 &  vrification !

		verifie_groupe1(node->move_coord.x,node->move_coord.y);

		if (!groupe1_vivant)
		{
			iprintf("suicide interdit! x=%d y=%d\n",node->move_coord.x,node->move_coord.y);
			iprintf("commande=%s",node->commande_SGF.c_str());
			return -1;
		}
		else
			return 0;
	}	
}

void CMySGFTree::Move_backward(Node_type* node)
{

	if (node->type_node==node_setup||node->type_node==node_neutral||node->type_node==node_pass)
	{
		//iprintf("Move_backward:setup!\n");		
		//WaitForTouch();
		return;
	}	
	
	// on remet la "capture_list" et on enlve la pierre "qui ne va plus bien" !
	Pierre_type pierre;
	if (node->move_couleur==PierreBlanche)
	{	pierre=PierreNoire;
	}
	else
	{	pierre=PierreBlanche;
	}

		if (node->move_couleur==PierreBlanche)
		{
			Total_prisonniers_noirs-=node->move_captures.size();
		}
		else
		{
			Total_prisonniers_blancs-=node->move_captures.size();
		}

	

	Uncapture_pierres(node->move_captures,pierre);



	pose_pierre(node->move_coord.x,node->move_coord.y,AucunePierre);

	// on pose le triangle

	/*Node_type* previous_node=node->node_amont;

	if (previous_node!=NULL)
	{
		pose_triangle_rouge(previous_node->move_coord.x,previous_node->move_coord.y);
	}	*/
}

void CMySGFTree::Capture_pierres(vector<Coordonnees_type>	move_captures)
{
	u16 x;

	for(x=0;x<move_captures.size();x++)
	{
		pose_pierre(move_captures[x].x,move_captures[x].y,AucunePierre);
	}
}

void CMySGFTree::Uncapture_pierres(vector<Coordonnees_type> move_captures,Pierre_type pierre)
{
	u16 x;

	for(x=0;x<move_captures.size();x++)
	{
		pose_pierre(move_captures[x].x,move_captures[x].y,pierre);
	}
}

u8 CMySGFTree::ExtractComment(string entete,vector<string> &commentaire)	// retourne le nombre
																			// de volets de 256*126
																			// ncessaires  l'affichage
{
	string::size_type p0,p1,p2;	

	p0=entete.find("C[",0);

	//iprintf("analyse de :%s",entete.c_str());
	//WaitForTouch();

	while (1)
	{					
		//iprintf("p0=%d;",p0);
		//WaitForTouch();
	
		if (p0!=string::npos && ((p0==0)||(entete[p0-1]==']')||(entete[p0-1]=='\n')))	// commentaire trouv!
		{
			p0+=2;
			p1=entete.find("]",p0);
			while (entete[p1-1]=='\\')	// juste pour viter les problmes avec des '\]' dans les attributs...
			{
				p1=entete.find("]",p1+1);
			}
			string full_comment=entete.substr(p0,p1-p0);
			
			//iprintf("tmp=/%s.\n",tmp.c_str());
			//WaitForTouch();
			
			u8 x=1;
			string reste,tmp;
			
			// on va maintenant dterminer combien de pages de 256*126 sont
			// ncessaires pour afficher le commentaire
			
			reste=arial12->WriteParagraph(full_comment,&page_commentaire,4,4,254,125);
			
			if (reste=="")
			{
				commentaire.push_back(full_comment);
				return 1;
			}
			else
			{				
				p2=full_comment.find(reste);
				if (p2==string::npos)
				{
					iprintf("Alors la, je ne comprends pas!!!\n");
					while(1);
				}
				
				tmp=full_comment.substr(0,p2);
				commentaire.push_back(tmp);	// et voil !
								
				while (reste!="")
				{			
					x++;
					commentaire.push_back(reste);
					reste=arial12->WriteParagraph(reste,&page_commentaire,4,4,254,125);
				}
				
				//iprintf("%d pages de commentaires\n",x);
				//WaitForTouch();
				
				return x;	// x pages de commentaires
				//break;
			}
		}
		if (p0==string::npos)
		{
			//iprintf("%d pages de commentaires\n",0);
			//WaitForTouch();
			// pas de commentaire trouv!			
			return 0;	// pas de volet ncessaire !
			//break;
		}
		p0+=2;		
		p0=entete.find("C[",p0);
	}
}

void CMySGFTree::RAZ()
{
	if (premiere_node==NULL)
		return;

	//iprintf("Welcome to RAZ!\n");

	//iprintf("allocated_goban_number=%d\n",allocated_goban_number);
	//iprintf("allocated_node_number=%d\n",allocated_node_number);
	//WaitForTouch();	

	// Nettoyage de l'arbre... destruction des nodes
	
	stack<Node_type*> pile;
	
	Node_type* current_node=premiere_node;
	
	// on empile chaque variante et on dtruit la node  l'origine...
	// quand la pile est vide, c'est termin !
	
	// on empile...
	
	while (current_node->variantes.size()>0)
	{
		//iprintf("on empile!\n");
		pile.push(current_node->variantes.back());
		current_node->variantes.pop_back();
	}	
	
	// on dtruit!
	
	if (current_node->setup_goban!=NULL)
	{
		//iprintf("on detruit le goban!\n");
		delete current_node->setup_goban;
		allocated_goban_number--;
	}
	//iprintf("on detruit la node!\n");
	delete current_node;
	allocated_node_number--;
	
	// Maintenant, tant qu'il en reste dans la pile, on recommence...
		
	while (pile.size()>0)
	{
		//iprintf("allocated_goban_number=%d\n",allocated_goban_number);
		//iprintf("allocated_node_number=%d\n",allocated_node_number);
		//WaitForTouch();
			
		current_node=pile.top();
		pile.pop();
		
		//iprintf("on dpile... size=%d\n",pile.size());
		
		while (current_node->variantes.size()>0)
		{			
			pile.push(current_node->variantes.back());
			current_node->variantes.pop_back();
			//iprintf("on empile! size=%d\n",pile.size());
		}	
		
		// on dtruit!
		
		if (current_node->setup_goban!=NULL)
		{
			//iprintf("on detruit le goban!\n");
			delete current_node->setup_goban;
			allocated_goban_number--;
		}
		//iprintf("on detruit la node!\n");
		delete current_node;
		allocated_node_number--;		
	}		
	
	// termin... normalement, les compteurs sont  zro...
	//iprintf("allocated_goban_number=%d\n",allocated_goban_number);
	//iprintf("allocated_node_number=%d\n",allocated_node_number);
	//WaitForTouch();
	
	premiere_node=NULL;
	
	// RAZ des attributs
	application="";
	charset="";
	fileformat="";
	game="";
	style="";
	size="";
	annotation="";	
	white_rank="";
	black_rank="";
	black_team="";
	white_team="";
	copyright="";
	date="";
	event="";
	game_comment="";
	game_name="";
	opening="";
	place="";
	result="";
	round="";
	rules="";
	source="";
	time_limit="";
	user="";
	handicap="";
	komi="";	
}

bool CMySGFTree::Enregistre_coup(Node_type* current_node, u8 xgoban,u8 ygoban,Pierre_type pierre)
{
	
	// enregistrement d'un nouveau coup...
	
	// on cre une nouvelle node pour recevoir les infos (quitte  la dtruire si le coup n'est pas valide)
	
	Node_type* nouvelle_node;
	
	nouvelle_node=new (Node_type);
	allocated_node_number++;
	
	nouvelle_node->setup_goban=NULL;
	nouvelle_node->move_number=current_node->move_number+1;
	nouvelle_node->nb_volets_commentaires=0;
	nouvelle_node->commande_SGF="";
	nouvelle_node->setup_goban=NULL;	
	nouvelle_node->node_amont=current_node;
	nouvelle_node->type_node=node_move;
	
	nouvelle_node->move_coord.x=xgoban;
	nouvelle_node->move_coord.y=ygoban;
	nouvelle_node->move_couleur=pierre;
	nouvelle_node->move_captures.clear();

	// nommage de la node...
	string texte;
	
	if (nouvelle_node->move_couleur==PierreBlanche)
		texte="W ";
	else
		texte="B ";
						
	texte+=gobanx2letter(nouvelle_node->move_coord.x)+to_string((u16)nouvelle_node->move_coord.y);
				
	nouvelle_node->node_name=texte;

	// est-ce que le coup est valide ?
	
	// test de capture...
				
	// on pose la pierre sur le goban...
				
	if (pierre==PierreBlanche)
	{
		Couleur_capturees=PierreNoire;
		Goban.intersection[xgoban][ygoban]=PierreBlanche;
	}
		else
	{
		Couleur_capturees=PierreBlanche;
		Goban.intersection[xgoban][ygoban]=PierreNoire;
	}
				
	// on vrifie s'il y a capture...
				
	verifie_capture(xgoban,ygoban);
				
	if ((Nb_capturees_g1+Nb_capturees_g2+Nb_capturees_g3+Nb_capturees_g4)>0)	// capture(s)!
	{
		// remplir la structure de capture
					
		// on remplit  la main groupe par groupe !
					
		u16 i,j;
		Coordonnees_type p;
		//p=node_courante->move_captures;
	
		for (i=1;i<=Goban.size;i++)
		{
			for (j=1;j<=Goban.size;j++)
			{
				if (repartition_groupes[i][j]==1 && Nb_capturees_g1>0)
				{
					p.x=i;
					p.y=j;
					nouvelle_node->move_captures.push_back(p);
				}
				else
				if (repartition_groupes[i][j]==2 && Nb_capturees_g2>0)
				{
					p.x=i;
					p.y=j;
					nouvelle_node->move_captures.push_back(p);
				}
				else
				if (repartition_groupes[i][j]==3 && Nb_capturees_g3>0)
				{
					p.x=i;
					p.y=j;
					nouvelle_node->move_captures.push_back(p);
				}
				else
				if (repartition_groupes[i][j]==4 && Nb_capturees_g4>0)
				{
					p.x=i;
					p.y=j;
					nouvelle_node->move_captures.push_back(p);
				}
			}					
		}
	}
	else
	{	
		// Maintenant : est-ce qu'il y a suicide ?
		// rappel : si capture => pas de suicide !
	
		u8 a,b;
		
		Couleur_capturees=pierre;
		
		Nb_capturees_g1=0;
		
		groupe1_vivant=false;	// jusqu' preuve du contraire !
		
		for (a=1;a<=19;a++)
			for (b=1;b<=19;b++)
			{
				repartition_groupes[a][b]=0;
			}
					
		// Affectation du groupe1 &  vrification !
		
		verifie_groupe1(xgoban,ygoban);
		
		if (!groupe1_vivant)
		{
			//iprintf("suicide interdit! x=%d y=%d\n",xgoban,ygoban);
			// retirer la pierre du goban
			PosePierreGobans(xgoban,ygoban,AucunePierre);
			// vide la liste de capture
			nouvelle_node->move_captures.clear();
			current_node->variantes.clear();
			delete nouvelle_node;
			allocated_node_number--;
			return false;
		}
	}
	
	// Le coup est valide...
	
	// On dtruit tout l'arbre en aval et on greffe la nouvelle node...
	
	if (current_node->variantes.size()>0)
	{
		// on dtruit les branches en aval de la node courante...
		Detruit_aval(current_node);
	}
	
	current_node->variantes.push_back(nouvelle_node);
	
	return true;

}

void CMySGFTree::Detruit_aval(Node_type* starting_node)
{
	if (starting_node==NULL)
		return;

	// Nettoyage de l'arbre... destruction des nodes
	
	stack<Node_type*> pile;
	
	Node_type* current_node=starting_node;
	
	// on empile chaque variante et on dtruit la node  l'origine...
	// quand la pile est vide, c'est termin !
	
	// on empile...
	
	while (current_node->variantes.size()>0)
	{
		//iprintf("on empile!\n");
		pile.push(current_node->variantes.back());
		current_node->variantes.pop_back();
	}	
	
	// Maintenant, tant qu'il en reste des branches, on dtruit...
		
	while (pile.size()>0)
	{
		current_node=pile.top();
		pile.pop();
		
		while (current_node->variantes.size()>0)
		{			
			pile.push(current_node->variantes.back());
			current_node->variantes.pop_back();
		}	
		
		// on dtruit!
		
		if (current_node->setup_goban!=NULL)
		{
			delete current_node->setup_goban;
			allocated_goban_number--;
		}
		
		delete current_node;
		allocated_node_number--;		
	}		
}
